import numpy as np
import utils
from sklearn.preprocessing import normalize, StandardScaler
import notears.notears.utils as notears_utils 
from notears.notears.linear import notears_linear
import argparse
from tqdm import tqdm
from pycausal.pycausal import pycausal as pc # pycausal package is used in fast greedy equivalence search algorithm (fGES or FGS)
from NoCurl.BPR import *
from NoCurl.main_efficient import *
import cdt 
cdt.SETTINGS.rpath = 'C:/Program Files/R/R-4.2.1/bin/Rscript'
pc = pc()
pc.start_vm()

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    # parser.add_argument('--value', default='variables', type=str, help='Choices variables, samples')
    parser.add_argument('--noise', default='gauss', type=str, help='Choices none, gauss, gumbel, uniform')
    parser.add_argument('--noise_std', default=0.01, type=float, help='Noise magnitude')
    parser.add_argument('--noise_effect', default='signal', type=str, help='Where the noise is applied ')
    parser.add_argument('--trans_clos', default='FW', type=str, help='Choices FW and zero')
    parser.add_argument('--sparsity', default=0.3, type=float, help='Probability of data being nonzero at vertex v')

    parser.add_argument('--weight_bounds', default=[0.4, 0.8], nargs='+', type=float, help='initialization of weighted adjacency matrix')
    parser.add_argument('--edges', default=2, type=int, help='graph has k * d edges')

    parser.add_argument('--samples', default=[400], nargs='+', type=int, help='number of samples')
    parser.add_argument('--nodes', default=[10, 20, 30, 40, 50], nargs='+', type=int, help='number of graph vertices to consider')# [5, 10, 15, 20, 25]
    parser.add_argument('--fixSup', default='False', type=str, help='Whether to fix the support of the spectrum')
    parser.add_argument('--methods', default=['moebius', 'notears', 'nocurl', 'FGS', 'GES', 'MMPC', 'CAM'], nargs='+', type=str, help='methods to compare')
    parser.add_argument('--transformation', default='None', type=str, help='Whether to normalize/standardize the given signals')
    parser.add_argument('--runs', default=10, type=int)
    args = parser.parse_args()

    print(vars(args))
    samples = args.samples 
    variables =  args.nodes 
    noise = args.noise
    runs = args.runs 
    (a, b) = tuple(args.weight_bounds)
    k = args.edges
    methods = args.methods
    fix_sup = (args.fixSup == 'True')

    # naming the output files according to the experimental settings 
    dic = vars(args)
    filename = ''
    label = ''
    for key in dic.keys():
        if(key not in ['methods', 'variables'] and dic[key]!= parser.get_default(key)):
            filename += '{}_{}_'.format(key, dic[key])

        label += '{} = {}, '.format(key, dic[key])
    filename = filename if len(filename) > 0 else 'default'


    with open('results/{}.csv'.format(filename), 'a') as f:
        for n in samples:
            for d in variables:
                f.write('{}\n'.format(label))
                print('samples = {}, variables = {}, noise = {}'.format(n, d, noise))

                current = {}

                for key in methods:
                    current[key] = []

                avg_cond_num = 0 
                for r in tqdm(range(runs)):
                    # graph initialization
                    B_true = notears_utils.simulate_dag(d, k * d, 'ER') # Erdös-Renyi graph simulation with avg degree = k
                    W_true = notears_utils.simulate_parameter(B_true, w_ranges=((-b, -a), (a, b))) # sampling uniformly the weights

                    # data initialization spectral SEM: X = C(I + \bar(W))
                    X, C_true, cond_num = utils.sparse_spectral_sem(W_true, n, sparsity=args.sparsity, noise_scale=args.noise_std, 
                                                                noise_type=noise, noise_effect=args.noise_effect, trans_clos=args.trans_clos, fix_sup=fix_sup)

                    # computing average condition number of (I + \bar(W))
                    avg_cond_num += cond_num

                    # applying transformation to data (or not)
                    if (args.transformation == 'norm'):
                        X = normalize(X)
                    elif (args.transformation == 'stand'):
                        scaler = StandardScaler().fit(X)
                        X = scaler.transform(X)
                
                    # causal discovery algorithms
                    for method in methods:
                        if method == 'moebius':
                            W_est = notears_linear(X, lambda1=0.001, loss_type='moebius', trans_clos=args.trans_clos)
                            B_est = W_est != 0
                        elif method == 'notears':
                            W_est = notears_linear(X, lambda1=0.005, loss_type='l2', trans_clos=args.trans_clos)
                            B_est = W_est != 0
                        else:
                            tmp_args = get_args()
                            bpr = BPR.BPR(tmp_args, pc)
                            W_est, h, alpha, rho = bpr.fit(X, method)
                            B_est = W_est != 0 if method == 'nocurl' else W_est

                        spectral_nmse, support_tpr = utils.spectrum_approximation(method, X, W_est, C_true, trans_clos=args.trans_clos)
                        nmse = np.linalg.norm(W_est - W_true) / np.linalg.norm(W_true)
                        acc = notears_utils.count_accuracy(B_true, B_est)
                        current[method].append([acc['shd'], acc['tpr'], acc['nnz'], nmse, support_tpr])

                # outputing cond num of matrices (I + transclos(W)) for all runs
                avg_cond_num = avg_cond_num / runs
                print('Avg cond num of (I + transclos(W)) is {:.3f}'.format(avg_cond_num))
                f.write('Avg cond num of (I + transclos(W)) is {:.3f}\n'.format(avg_cond_num))

                # Log results
                avg = {}
                std = {}

                for method in methods:
                    avg[method] = np.mean(current[method], axis=0)
                    std[method] = np.std(current[method], axis=0)
                    f.write("Acc {} is, {:.3f}, {:.3f}, {:.3f}, {:.3f}, {:.3f}\n".format(method, avg[method][0], avg[method][1], avg[method][2], avg[method][3], avg[method][4]))
                    f.write("Std {} is, {:.3f}, {:.3f}, {:.3f}, {:.3f}, {:.3f}\n".format(method, std[method][0], std[method][1], std[method][2], std[method][3], std[method][4]))
                    print("Acc {} is, {:.3f}, {:.3f}, {:.3f}, {:.3f}, {:.3f}".format(method, avg[method][0], avg[method][1], avg[method][2], avg[method][3], avg[method][4]))
                    print("Std {} is, {:.3f}, {:.3f}, {:.3f}, {:.3f}, {:.3f}".format(method, std[method][0], std[method][1], std[method][2], std[method][3], std[method][4]))

pc.stop_vm()