import numpy as np

def sparse_spectral_sem(W, n, sparsity=0.3, noise_scale=None, noise_type='gauss', noise_effect='signal', trans_clos='FW', std=False, fix_sup=False):
    #number of nodes
    d = W.shape[0]

    # initializing the sparse spectrum
    if(fix_sup):
        pos = np.random.choice([0, 1], size=d, p=[1 - sparsity, sparsity])
    else: 
        pos = np.random.choice([0, 1], size=(n, d), p=[1 - sparsity, sparsity])
    C = pos * np.random.uniform(0, 1, size=(n, d)) 

    # computing matrix of indipendent noises
    if noise_scale is None:
        noise = np.zeros((n, d))
    elif noise_type == 'gauss':
        noise = np.random.normal(scale=noise_scale, size=(n, d))
    else: # considering gumbel case
        if(std):
            noise_scale = np.sqrt(6) * noise_scale / np.pi
        noise = np.random.gumbel(scale=noise_scale, size=(n, d))

    # computing \bar(W) according to definition of transitive closure (either Floyd-Warshall or zero)
    refl_trans_clos = np.linalg.inv(np.eye(d) - W) if trans_clos == 'FW' else (np.eye(d) + W)
     
    if noise_effect == 'spectral':
        # computing X = (C + N)(I + \bar(W))
        X = (C + noise) @ refl_trans_clos 
    elif noise_effect == 'signal':
        # computing X = C(I + \bar(W)) + N
        X = C @ refl_trans_clos + noise

    # Condition number of (I + transclos(W)) matrix
    cond_num = np.linalg.cond(refl_trans_clos)

    return X, C, cond_num


def spectrum_approximation(method, X, W_est, C_true, trans_clos='FW', epsilon=0.1):
    '''
    naive implementation of approximating C in the cases:
    X = (C + N)(I + \bar(W))
    and 
    C(I + \bar(W)) + N

    returns nan or spectral nmse
    '''
    if method not in ['moebius', 'notears', 'nocurl']:
        return float("nan"), float("nan") # not applicable to the algorithms that return unweighted adjacency matrix
    else:
        d = W_est.shape[0]
        inverse_refl_trans_clos = np.eye(d) - W_est if trans_clos == 'FW' else np.linalg.inv(np.eye(d) + W_est)
        C_est = X @ inverse_refl_trans_clos
        spectral_nmse = np.linalg.norm(C_est - C_true) / np.linalg.norm(C_true)

        non_zero_est = np.where(C_est > epsilon * np.max(C_est), 1, 0)
        non_zero_true = np.where(C_true > epsilon * np.max(C_true), 1, 0)
        spectral_support_tpr = np.sum(non_zero_est * non_zero_true) / np.sum(non_zero_true)

    return spectral_nmse, spectral_support_tpr
